import json
import logging
import logging.handlers
import requests
import socket
import sys
import datetime
import time

from .amsexceptions import (AmsServiceException, AmsConnectionException,
                            AmsMessageException, AmsException,
                            AmsTimeoutException, AmsBalancerException)
from .amsmsg import AmsMessage
from .amstopic import AmsTopic
from .amssubscription import AmsSubscription
from .amsuser import AmsUser

try:
    from collections import OrderedDict
except:
    from ordereddict import OrderedDict

log = logging.getLogger(__name__)


class AmsHttpRequests(object):
    """Class encapsulates methods used by ArgoMessagingService.

       Each method represent HTTP request made to AMS with the help of requests
       library. service error handling is implemented according to HTTP
       status codes returned by service and the balancer.
    """

    def __init__(self, endpoint, authn_port, token="", cert="", key=""):
        self.endpoint = endpoint
        self.authn_port = authn_port
        self.token = token

        # Create route list
        self.routes = {
            # topic api calls
            "topic_list": ["get", "https://{0}/v1/projects/{1}/topics"],
            "topic_get": ["get", "https://{0}/v1/projects/{1}/topics/{2}"],
            "topic_publish": ["post", "https://{0}/v1/projects/{1}/topics/{2}:publish"],
            "topic_create": ["put", "https://{0}/v1/projects/{1}/topics/{2}"],
            "topic_delete": ["delete", "https://{0}/v1/projects/{1}/topics/{2}"],
            "topic_getacl": ["get", "https://{0}/v1/projects/{1}/topics/{2}:acl"],
            "topic_modifyacl": ["post", "https://{0}/v1/projects/{1}/topics/{2}:modifyAcl"],

            # subscription api calls
            "sub_create": ["put", "https://{0}/v1/projects/{1}/subscriptions/{2}"],
            "sub_delete": ["delete", "https://{0}/v1/projects/{1}/subscriptions/{2}"],
            "sub_list": ["get", "https://{0}/v1/projects/{1}/subscriptions"],
            "sub_get": ["get", "https://{0}/v1/projects/{1}/subscriptions/{2}"],
            "sub_pull": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:pull"],
            "sub_ack": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:acknowledge"],
            "sub_pushconfig": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:modifyPushConfig"],
            "sub_getacl": ["get", "https://{0}/v1/projects/{1}/subscriptions/{2}:acl"],
            "sub_modifyacl": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:modifyAcl"],
            "sub_offsets": ["get", "https://{0}/v1/projects/{1}/subscriptions/{2}:offsets"],
            "sub_mod_offset": ["post", "https://{0}/v1/projects/{1}/subscriptions/{2}:modifyOffset"],
            "sub_timeToOffset": ["get", "https://{0}/v1/projects/{1}/subscriptions/{2}:timeToOffset?time={3}"],

            # user api calls
            "user_create": ["post", "https://{0}/v1/users/{1}"],
            "user_get": ["get", "https://{0}/v1/users/{1}"],

            # project api calls
            "project_add_member": ["post", "https://{0}/v1/projects/{1}/members/{2}:add"],
            "project_get_member": ["get", "https://{0}/v1/projects/{1}/members/{2}"],

            "auth_x509": ["get", "https://{0}:{1}/v1/service-types/ams/hosts/{0}:authx509"],
        }

        # HTTP error status codes returned by AMS according to:
        # http://argoeu.github.io/messaging/v1/api_errors/
        self.ams_errors_route = {
            "topic_create": ["put", set([409, 401, 403])],
            "topic_list": ["get", set([400, 401, 403, 404])],
            "topic_delete": ["delete", set([401, 403, 404])],
            "topic_get": ["get", set([404, 401, 403])],
            "topic_modifyacl": ["post", set([400, 401, 403, 404])],
            "topic_publish": ["post", set([413, 401, 403])],

            "sub_create": ["put", set([400, 409, 408, 401, 403])],
            "sub_get": ["get", set([404, 401, 403])],
            "sub_mod_offset": ["post", set([400, 401, 403, 404])],
            "sub_ack": ["post", set([408, 400, 401, 403, 404])],
            "sub_pushconfig": ["post", set([400, 401, 403, 404])],
            "sub_pull": ["post", set([400, 401, 403, 404])],
            "sub_timeToOffset": ["get", set([400, 401, 403, 404, 409])],

            "user_create": ["post", set([400, 401, 403, 404, 409])],
            "user_get": ["post", set([400, 401, 403, 404])],

            "project_add_member": ["post", set([400, 401, 403, 404, 409])],
            "project_get_member": ["get", set([400, 401, 403, 404])],

            "auth_x509": ["post", set([400, 401, 403, 404])]}

        # https://cbonte.github.io/haproxy-dconv/1.8/configuration.html#1.3
        self.balancer_errors_route = {"sub_ack": ["post", set([500, 502, 503, 504])],
                                      "sub_pull": ["post", set([500, 502, 503, 504])],
                                      "topic_publish": ["post", set([500, 502, 503, 504])]}

        # determine the token to be used
        self.assign_token(token, cert, key)

    def assign_token(self, token, cert, key):
        """Assign a token to the ams object

           Args:
               token(str): a valid ams token
               cert(str): a path to a valid certificate file
               key(str): a path to the associated key file for the provided certificate
        """

        # check if a token has been provided
        if token != "":
            return

        try:
            # otherwise use the provided certificate to retrieve it
            self.token = self.auth_via_cert(cert, key)
        except AmsServiceException as e:
            # if the request send to authn didn't contain an x509 cert, that means that there was also no token provided
            # when initializing the ArgoMessagingService object, since we only try to authenticate through authn
            # when no token was provided

            if e.msg == 'While trying the [auth_x509]: No certificate provided.':
                refined_msg = "No certificate provided. No token provided."
                errormsg = self._error_dict(refined_msg, e.code)
                raise AmsServiceException(json=errormsg, request="auth_x509")
            raise e

    def auth_via_cert(self, cert, key, **reqkwargs):
        """Retrieve an ams token based on the provided certificate

            Args:
                cert(str): a path to a valid certificate file
                key(str): a path to the associated key file for the provided certificate

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        if cert == "" and key == "":
            errord = self._error_dict("No certificate provided.", 400)
            raise AmsServiceException(json=errord, request="auth_x509")

        # create the certificate tuple needed by the requests library
        reqkwargs = {"cert": (cert, key)}

        route = self.routes["auth_x509"]

        # Compose url
        url = route[1].format(self.endpoint, self.authn_port)
        method = getattr(self, 'do_{0}'.format(route[0]))

        try:
            r = method(url, "auth_x509", **reqkwargs)
            # if the `token` field was not found in the response, raise an error
            if "token" not in r:
                errord = self._error_dict("Token was not found in the response. Response: " + str(r), 500)
                raise AmsServiceException(json=errord, request="auth_x509")
            return r["token"]
        except (AmsServiceException, AmsConnectionException) as e:
            raise e

    def _error_dict(self, response_content, status):
        error_dict = dict()

        try:
            if (response_content and sys.version_info < (3, 6,) and
                    isinstance(response_content, bytes)):
                response_content = response_content.decode()
            error_dict = json.loads(response_content) if response_content else {}
        except ValueError:
            error_dict = {'error': {'code': status, 'message': response_content}}

        return error_dict

    def _gen_backoff_time(self, try_number, backoff_factor):
        for i in range(0, try_number):
            value = backoff_factor * (2 ** (i - 1))
            yield value

    def _retry_make_request(self, url, body=None, route_name=None, retry=0,
                            retrysleep=60, retrybackoff=None, **reqkwargs):
        """Wrapper around _make_request() that decides whether should request
        be retried or not.

           Two request retry modes are available:
               1) static sleep - fixed amount of seconds to sleep between
                  request attempts
               2) backoff - each next sleep before request attempt is
                  exponentially longer

           If enabled, request will be retried in the following occassions:
               * timeouts from AMS (HTTP 408) or load balancer (HTTP 408 and 504)
               * load balancer HTTP 502, 503
               * connection related problems in the lower network layers

           Default behaviour is no retry attempts. If both, retry and
           retrybackoff are enabled, retrybackoff will take precedence.

           Args:
               url: str. The final messaging service endpoint
               body: dict. Payload of the request
               route_name: str. The name of the route to follow selected from the route list
               retry: int. Number of request retries before giving up. Default
                           is 0 meaning no further request retry will be made
                           after first unsuccesfull request.
               retrysleep: int. Static number of seconds to sleep before next
                           request attempt
               retrybackoff: int. Backoff factor to apply between each request
                             attempts
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        i = 1
        timeout = reqkwargs.get('timeout', 0)

        saved_exp = None
        if retrybackoff:
            try:
                return self._make_request(url, body, route_name, **reqkwargs)
            except (AmsBalancerException, AmsConnectionException,
                    AmsTimeoutException) as e:
                for sleep_secs in self._gen_backoff_time(retry, retrybackoff):
                    try:
                        return self._make_request(url, body, route_name, **reqkwargs)
                    except (AmsBalancerException, AmsConnectionException,
                            AmsTimeoutException) as e:
                        saved_exp = e
                        time.sleep(sleep_secs)
                        if timeout:
                            log.warning(
                                'Backoff retry #{0} after {1} seconds, connection timeout set to {2} seconds - {3}: {4}'.format(
                                    i, sleep_secs, timeout, self.endpoint, e))
                        else:
                            log.warning(
                                'Backoff retry #{0} after {1} seconds - {2}: {3}'.format(i, sleep_secs, self.endpoint,
                                                                                         e))
                    finally:
                        i += 1
                else:
                    if saved_exp:
                        raise saved_exp
                    else:
                        raise e

        else:
            while i <= retry + 1:
                try:
                    return self._make_request(url, body, route_name, **reqkwargs)
                except (AmsBalancerException, AmsConnectionException, AmsTimeoutException) as e:
                    if i == retry + 1:
                        raise e
                    else:
                        time.sleep(retrysleep)
                        if timeout:
                            log.warning(
                                'Retry #{0} after {1} seconds, connection timeout set to {2} seconds - {3}: {4}'.format(
                                    i, retrysleep, timeout, self.endpoint, e))
                        else:
                            log.warning(
                                'Retry #{0} after {1} seconds - {2}: {3}'.format(i, retrysleep, self.endpoint, e))
                finally:
                    i += 1

    def _make_request(self, url, body=None, route_name=None, **reqkwargs):
        """Common method for PUT, GET, POST HTTP requests with appropriate
           service error handling by differing between AMS and load balancer
           erroneous behaviour.
        """
        m = self.routes[route_name][0]
        decoded = None
        try:
            # the get request based on requests.

            # populate all requests with the x-api-key header
            # except the authn mapping call
            if route_name != "auth_x509":
                # if there is no defined headers dict in the reqkwargs, introduce it
                if "headers" not in reqkwargs:
                    headers = {
                        "x-api-key": self.token
                    }
                    reqkwargs["headers"] = headers
                else:
                    # if the there are already other headers defined, just append the x-api-key one
                    reqkwargs["headers"]["x-api-key"] = self.token

            reqmethod = getattr(requests, m)
            r = reqmethod(url, data=body, **reqkwargs)

            content = r.content
            status_code = r.status_code

            if (content and sys.version_info < (3, 6,) and isinstance(content,
                                                                      bytes)):
                content = content.decode()

            if status_code == 200:
                decoded = self._error_dict(content, status_code)

            # handle authnz related errors for all calls
            elif status_code == 401 or status_code == 403:
                raise AmsServiceException(json=self._error_dict(content,
                                                                status_code),
                                          request=route_name)

            elif status_code == 408 or (status_code == 504 and route_name in
                                        self.balancer_errors_route):
                raise AmsTimeoutException(json=self._error_dict(content,
                                                                status_code),
                                          request=route_name)

            # handle errors from AMS
            elif (status_code != 200 and status_code in
                  self.ams_errors_route[route_name][1]):
                raise AmsServiceException(json=self._error_dict(content,
                                                                status_code),
                                          request=route_name)

            # handle errors coming from load balancer
            elif (status_code != 200 and route_name in
                  self.balancer_errors_route and status_code in
                  self.balancer_errors_route[route_name][1]):
                raise AmsBalancerException(json=self._error_dict(content,
                                                                 status_code),
                                           request=route_name)

            # handle any other erroneous behaviour by raising exception
            else:
                raise AmsServiceException(json=self._error_dict(content,
                                                                status_code),
                                          request=route_name)

        except (requests.exceptions.ConnectionError, socket.error) as e:
            raise AmsConnectionException(e, route_name)

        else:
            return decoded if decoded else {}

    def do_get(self, url, route_name, **reqkwargs):
        """Method supports all the GET requests.

           Used for (topics, subscriptions, users, messages).

           Args:
               url: str. The final messaging service endpoint
               route_name: str. The name of the route to follow selected from the route list
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        # try to send a GET request to the messaging service.
        # if a connection problem araises a Connection error exception is raised.
        try:
            return self._retry_make_request(url, route_name=route_name,
                                            **reqkwargs)
        except AmsException as e:
            raise e

    def do_put(self, url, body, route_name, **reqkwargs):
        """Method supports all the PUT requests.

           Used for (topics, subscriptions, messages).

           Args:
               url: str. The final messaging service endpoint
               body: dict. Body the post data to send based on the PUT request.
                           The post data is always in json format.
               route_name: str. The name of the route to follow selected from
                           the route list
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        # try to send a PUT request to the messaging service.
        # if a connection problem arises a Connection error exception is raised.
        try:
            return self._retry_make_request(url, body=body,
                                            route_name=route_name, **reqkwargs)
        except AmsException as e:
            raise e

    def do_post(self, url, body, route_name, retry=0, retrysleep=60,
                retrybackoff=None, **reqkwargs):
        """Method supports all the POST requests.

           Used for (topics, subscriptions, users, messages).

           Args:
               url: str. The final messaging service endpoint
               body: dict. Body the post data to send based on the PUT request.
                     The post data is always in json format.
               route_name: str. The name of the route to follow selected from
                           the route list
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        # try to send a Post request to the messaging service.
        # if a connection problem araises a Connection error exception is raised.
        try:
            return self._retry_make_request(url, body=body,
                                            route_name=route_name, retry=retry,
                                            retrysleep=retrysleep,
                                            retrybackoff=retrybackoff,
                                            **reqkwargs)
        except AmsException as e:
            raise e

    def do_delete(self, url, route_name, **reqkwargs):
        """Delete method that is used to make the appropriate request.

           Used for (topics, subscriptions).

           Args:
               url: str. The final messaging service endpoint
               route_name: str. The name of the route to follow selected from the route list
               reqkwargs: keyword argument that will be passed to underlying python-requests library call.
        """
        # try to send a delete request to the messaging service.
        # if a connection problem araises a Connection error exception is raised.
        try:
            # the delete request based on requests.
            r = requests.delete(url, **reqkwargs)

            # JSON error returned by AMS
            if r.status_code != 200:
                errormsg = self._error_dict(r.content, r.status_code)
                raise AmsServiceException(json=errormsg, request=route_name)

            else:
                return True

        except (requests.exceptions.ConnectionError,
                requests.exceptions.Timeout) as e:
            raise AmsConnectionException(e, route_name)


class ArgoMessagingService(AmsHttpRequests):
    """Class is entry point for client code.

       Class abstract Argo Messaging Service by covering all available HTTP API
       calls that are wrapped in series of methods.
    """

    def __init__(self, endpoint, token="", project="", cert="", key="", authn_port=8443):
        super(ArgoMessagingService, self).__init__(endpoint, authn_port, token, cert, key)
        self.project = project
        self.pullopts = {"maxMessages": "1",
                         "returnImmediately": "false"}
        # Containers for topic and subscription objects
        self.topics = OrderedDict()
        self.subs = OrderedDict()

    def _create_sub_obj(self, s, topic):
        self.subs.update({s['name']: AmsSubscription(s['name'], topic,
                                                     s['pushConfig'],
                                                     s['ackDeadlineSeconds'],
                                                     init=self)})

    def _delete_sub_obj(self, s):
        del self.subs[s['name']]

    def _create_topic_obj(self, t):
        self.topics.update({t['name']: AmsTopic(t['name'], init=self)})

    def _delete_topic_obj(self, t):
        del self.topics[t['name']]

    def getacl_topic(self, topic, **reqkwargs):
        """Get access control lists for topic

           Args:
               topic (str): The topic name.

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """

        topicobj = self.get_topic(topic, retobj=True)

        route = self.routes["topic_getacl"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, topic)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, "topic_getacl", **reqkwargs)

        if r:
            self.topics[topicobj.fullname].acls = r['authorized_users']
            return r
        else:
            self.topics[topicobj.fullname].acls = []
            return []

    def modifyacl_topic(self, topic, users, **reqkwargs):
        """Modify access control lists for topic

           Args:
               topic (str): The topic name.
               users (list): List of users that will have access to topic.
                             Empty list of users will reset access control list.

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """

        topicobj = self.get_topic(topic, retobj=True)

        route = self.routes["topic_modifyacl"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, topic)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = None
        try:
            msg_body = json.dumps({"authorized_users": users})
            r = method(url, msg_body, "topic_modifyacl", **reqkwargs)

            if r is not None:
                self.topics[topicobj.fullname].acls = users

            return True

        except (AmsServiceException, AmsConnectionException) as e:
            raise e

    def getacl_sub(self, sub, **reqkwargs):
        """Get access control lists for subscription

           Args:
               sub (str): The subscription name.

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """

        subobj = self.get_sub(sub, retobj=True)

        route = self.routes["sub_getacl"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, "sub_getacl", **reqkwargs)

        if r:
            self.subs[subobj.fullname].acls = r['authorized_users']
            return r
        else:
            self.subs[subobj.fullname].acls = []
            return []

    def getoffsets_sub(self, sub, offset='all', **reqkwargs):
        """Retrieve the current positions of min,max and current offsets.

           Args:
               sub (str): The subscription name.
               offset(str): The name of the offset.If not specified, it will return all three of them as a dict.

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["sub_offsets"]

        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))
        r = method(url, "sub_offsets", **reqkwargs)
        try:
            if offset != 'all':
                return r[offset]
            return r
        except KeyError as e:
            errormsg = {'error': {'message': str(e) + " is not valid offset position"}}
            raise AmsServiceException(json=errormsg, request="sub_offsets")

    def time_to_offset_sub(self, sub, timestamp, **reqkwargs):
        """Retrieve the closest(greater than) available offset to the given timestamp.

           Args:
               sub (str): The subscription name.
               timestamp(datetime.datetime): The timestamp of the offset we are looking for.

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["sub_timeToOffset"]
        method = getattr(self, 'do_{0}'.format(route[0]))

        time_in_string = ""

        if isinstance(timestamp, datetime.datetime):
            if timestamp.microsecond != 0:
                time_in_string = timestamp.isoformat()[:-3] + "Z"
            else:
                time_in_string = timestamp.strftime("%Y-%m-%d %H:%M:%S.000Z")

        # Compose url
        url = route[1].format(self.endpoint, self.project, sub, time_in_string)

        try:
            r = method(url, "sub_timeToOffset", **reqkwargs)
            return r["offset"]
        except AmsServiceException as e:
            raise e

    def modifyoffset_sub(self, sub, move_to, **reqkwargs):
        """Modify the position of the current offset.

           Args:
               sub (str): The subscription name.
               move_to(int): Position to move the offset.

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["sub_mod_offset"]
        method = getattr(self, 'do_{0}'.format(route[0]))

        if not isinstance(move_to, int):
            move_to = int(move_to)

        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)

        # Request body
        data = {"offset": move_to}
        try:
            r = method(url, json.dumps(data), "sub_mod_offset", **reqkwargs)
            return r
        except AmsServiceException as e:
            raise e

    def modifyacl_sub(self, sub, users, **reqkwargs):
        """Modify access control lists for subscription

           Args:
               sub (str): The subscription name.
               users (list): List of users that will have access to subscription.
                             Empty list of users will reset access control list.
           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """

        subobj = self.get_sub(sub, retobj=True)

        route = self.routes["sub_modifyacl"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = None
        try:
            msg_body = json.dumps({"authorized_users": users})
            r = method(url, msg_body, "sub_modifyacl", **reqkwargs)

            if r is not None:
                self.subs[subobj.fullname].acls = users

            return True

        except (AmsServiceException, AmsConnectionException) as e:
            raise e

    def pushconfig_sub(self, sub, push_endpoint=None, retry_policy_type='linear', retry_policy_period=300, **reqkwargs):
        """Modify push configuration of given subscription

           Args:
               sub: shortname of subscription
               push_endpoint: URL of remote endpoint that should receive messages in push subscription mode
               retry_policy_type:
               retry_policy_period:
               reqkwargs: keyword argument that will be passed to underlying python-requests library call.
        """
        if push_endpoint:
            push_dict = {"pushConfig": {"pushEndpoint": push_endpoint,
                                        "retryPolicy": {"type": retry_policy_type,
                                                        "period": retry_policy_period}}}
        else:
            push_dict = {"pushConfig": {}}

        msg_body = json.dumps(push_dict)
        route = self.routes["sub_pushconfig"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))
        p = method(url, msg_body, "sub_pushconfig", **reqkwargs)

        subobj = self.subs.get('/projects/{0}/subscriptions/{1}'.format(self.project, sub), False)
        if subobj:
            subobj.push_endpoint = push_endpoint
            subobj.retry_policy_type = retry_policy_type
            subobj.retry_policy_period = retry_policy_period

        return p

    def iter_subs(self, topic=None, **reqkwargs):
        """Iterate over AmsSubscription objects

           Args:
               topic: Iterate over subscriptions only associated to this topic name
        """
        self.list_subs(**reqkwargs)

        try:
            values = self.subs.copy().itervalues()
        except AttributeError:
            values = self.subs.copy().values()

        for s in values:
            if topic and topic == s.topic.name:
                yield s
            elif not topic:
                yield s

    def iter_topics(self, **reqkwargs):
        """Iterate over AmsTopic objects"""

        self.list_topics(**reqkwargs)

        try:
            values = self.topics.copy().itervalues()
        except AttributeError:
            values = self.topics.copy().values()

        for t in values:
            yield t

    def list_topics(self, **reqkwargs):
        """List the topics of a selected project

           Args:
               reqkwargs: keyword argument that will be passed to underlying
               python-requests library call
        """
        route = self.routes["topic_list"]
        # Compose url
        url = route[1].format(self.endpoint, self.project)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, "topic_list", **reqkwargs)

        for t in r['topics']:
            if t['name'] not in self.topics:
                self._create_topic_obj(t)

        if r:
            return r
        else:
            return []

    def has_topic(self, topic, **reqkwargs):
        """Inspect if topic already exists or not

           Args:
               topic: str. Topic name
        """
        try:
            self.get_topic(topic, **reqkwargs)
            return True

        except AmsServiceException as e:
            if e.code == 404:
                return False
            else:
                raise e

        except AmsConnectionException as e:
            raise e

    def get_topic(self, topic, retobj=False, **reqkwargs):
        """Get the details of a selected topic.

           Args:
               topic: str. Topic name.
               retobj: Controls whether method should return AmsTopic object
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["topic_get"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, topic)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, "topic_get", **reqkwargs)

        if r['name'] not in self.topics:
            self._create_topic_obj(r)

        if retobj:
            return self.topics[r['name']]
        else:
            return r

    def publish(self, topic, msg, retry=0, retrysleep=60, retrybackoff=None, **reqkwargs):
        """Publish a message or list of messages to a selected topic.

           If enabled (retry > 0), multiple topic publishes will be tried in
           case of problems/glitches with the AMS service. retry* options are
           eventually passed to _retry_make_request()

           Args:
               topic (str): Topic name.
               msg (list): A list with one or more messages to send.
                           Each message is represented as AmsMessage object or python
                           dictionary with at least data or one attribute key defined.
           Kwargs:
               retry: int. Number of request retries before giving up. Default
                           is 0 meaning no further request retry will be made
                           after first unsuccesfull request.
               retrysleep: int. Static number of seconds to sleep before next
                           request attempt
               retrybackoff: int. Backoff factor to apply between each request
                             attempts
               reqkwargs: keyword argument that will be passed to underlying
                       python-requests library call.
           Return:
               dict: Dictionary with messageIds of published messages
        """
        if not isinstance(msg, list):
            msg = [msg]
        if all(isinstance(m, AmsMessage) for m in msg):
            msg = [m.dict() for m in msg]
        try:
            msg_body = json.dumps({"messages": msg})
        except TypeError as e:
            raise AmsMessageException(e)

        route = self.routes["topic_publish"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, topic)
        method = getattr(self, 'do_{0}'.format(route[0]))

        return method(url, msg_body, "topic_publish", retry=retry,
                      retrysleep=retrysleep, retrybackoff=retrybackoff,
                      **reqkwargs)

    def list_subs(self, **reqkwargs):
        """Lists all subscriptions in a project with a GET request.

           Args:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["sub_list"]
        # Compose url
        url = route[1].format(self.endpoint, self.project)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, "sub_list", **reqkwargs)

        for s in r['subscriptions']:
            if s['topic'] not in self.topics:
                self._create_topic_obj({'name': s['topic']})
            if s['name'] not in self.subs:
                self._create_sub_obj(s, self.topics[s['topic']].fullname)

        if r:
            return r
        else:
            return []

    def get_sub(self, sub, retobj=False, **reqkwargs):
        """Get the details of a subscription.

           Args:
               sub: str. The subscription name.
               retobj: Controls whether method should return AmsSubscription object
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["sub_get"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, "sub_get", **reqkwargs)

        if r['topic'] not in self.topics:
            self._create_topic_obj({'name': r['topic']})
        if r['name'] not in self.subs:
            self._create_sub_obj(r, self.topics[r['topic']].fullname)

        if retobj:
            return self.subs[r['name']]
        else:
            return r

    def has_sub(self, sub, **reqkwargs):
        """Inspect if subscription already exists or not

           Args:
               sub: str. The subscription name.
        """
        try:
            self.get_sub(sub, **reqkwargs)
            return True

        except AmsServiceException as e:
            if e.code == 404:
                return False
            else:
                raise e

        except AmsConnectionException as e:
            raise e

    def pull_sub(self, sub, num=1, return_immediately=False, retry=0,
                 retrysleep=60, retrybackoff=None, **reqkwargs):
        """This function consumes messages from a subscription in a project
           with a POST request.

           If enabled (retry > 0), multiple subscription pulls will be tried in
           case of problems/glitches with the AMS service. retry* options are
           eventually passed to _retry_make_request()

           Args:
               sub: str. The subscription name.
               num: int. The number of messages to pull.
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """

        wasmax = self.get_pullopt('maxMessages')
        wasretim = self.get_pullopt('returnImmediately')

        self.set_pullopt('maxMessages', num)
        self.set_pullopt('returnImmediately', str(return_immediately).lower())
        msg_body = json.dumps(self.pullopts)

        route = self.routes["sub_pull"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))
        r = method(url, msg_body, "sub_pull", retry=retry,
                   retrysleep=retrysleep, retrybackoff=retrybackoff,
                   **reqkwargs)
        msgs = r['receivedMessages']

        self.set_pullopt('maxMessages', wasmax)
        self.set_pullopt('returnImmediately', wasretim)

        return list(map(lambda m: (m['ackId'], AmsMessage(b64enc=False, **m['message'])), msgs))

    def ack_sub(self, sub, ids, **reqkwargs):
        """Acknownledgment of received messages

           Messages retrieved from a pull subscription can be acknowledged by
           sending message with an array of ackIDs. The service will retrieve
           the ackID corresponding to the highest message offset and will
           consider that message and all previous messages as acknowledged by
           the consumer.

           Args:
              sub: str. The subscription name.
              ids: list(str). A list of ids of the messages to acknowledge.
              reqkwargs: keyword argument that will be passed to underlying python-requests library call.
        """

        msg_body = json.dumps({"ackIds": ids})

        route = self.routes["sub_ack"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))
        method(url, msg_body, "sub_ack", **reqkwargs)

        return True

    def pullack_sub(self, sub, num=1, return_immediately=False, retry=0,
                    retrysleep=60, retrybackoff=None, **reqkwargs):
        """Pull messages from subscription and acknownledge them in one call.

           If enabled (retry > 0), multiple subscription pulls will be tried in
           case of problems/glitches with the AMS service. retry* options are
           eventually passed to _retry_make_request().

           If succesfull subscription pull immediately follows with failed
           acknownledgment (e.g. network hiccup just before acknowledgement of
           received messages), consume cycle will reset and start from
           beginning with new subscription pull. This ensures that ack deadline
           time window is moved to new start period, that is the time when the
           second pull was initiated.

           Args:
               sub: str. The subscription name.
               num: int. The number of messages to pull.
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        while True:
            try:
                ackIds = list()
                messages = list()

                for id, msg in self.pull_sub(sub, num,
                                             return_immediately=return_immediately,
                                             retry=retry,
                                             retrysleep=retrysleep,
                                             retrybackoff=retrybackoff,
                                             **reqkwargs):
                    ackIds.append(id)
                    messages.append(msg)

            except AmsException as e:
                raise e

            if messages and ackIds:
                try:
                    self.ack_sub(sub, ackIds, **reqkwargs)
                    break
                except AmsException as e:
                    log.warning('Continuing with sub_pull after sub_ack: {0}'.format(e))
                    pass
            else:
                break

        return messages

    def set_pullopt(self, key, value):
        """Function for setting pull options

           Args:
               key: str. The name of the pull option (ex. maxMessages, returnImmediately). Messaging specific
                    names are allowed.
               value: str or int. The name of the pull option (ex. maxMessages,
                      returnImmediately). Messaging specific names are allowed.
        """

        self.pullopts.update({key: str(value)})

    def get_pullopt(self, key):
        """Function for getting pull options

           Args:
               key: str. The name of the pull option (ex. maxMessages,
                    returnImmediately). Messaging specific names are allowed.

           Returns:
               str. The value of the pull option
        """
        return self.pullopts[key]

    def create_user(self, user, **reqkwargs):

        """This function creates a new user with a POST request

           Args:
               user: AmsUser. The user to be created.
               reqkwargs: keyword argument that will be passed to underlying
               python-requests library call.
           Return:
               object (AmsUser)
        """

        if not isinstance(user, AmsUser):
            raise ValueError("user has to be of type AmsUser")

        try:
            route = self.routes["user_create"]
            url = route[1].format(self.endpoint, user.name)
            method = getattr(self, 'do_{0}'.format(route[0]))
            r = method(url, user.to_json(), "user_create", **reqkwargs)
            return AmsUser().load_from_dict(r)
        except AmsException as e:
            raise e

    def get_user(self, name, **reqkwargs):
        """
        Retrieves the respective user using the provided username with a GET request

        :param name: (str) the username of the user to be retrieved
        :param reqkwargs:  keyword arguments that will be passed to underlying
               python-requests library call.
        :return: (AmsUser) the ams user
        """
        try:
            route = self.routes["user_get"]
            url = route[1].format(self.endpoint, name)
            method = getattr(self, 'do_{0}'.format(route[0]))
            r = method(url, "user_get", **reqkwargs)
            return AmsUser().load_from_dict(r)
        except AmsException as e:
            raise e

    def add_project_member(self, username, project=None, roles=None, **reqkwargs):
        """
        Assigns an existing user to the provided project with a POST request

        :param (str) project: the name of the project.If no
        project is supplied, the declared global project will be used instead
        :param (str) username: the name of user
        :param (str[]) roles: project roles for the user
        :param reqkwargs:  keyword argument that will be passed to underlying
                          python-requests library call.
        :return: (AmsUser) the assigned user object
        """

        if roles is None or not isinstance(roles, list):
            roles = []

        if project is None:
            project = self.project

        body = {
                "project": project,
                "roles": roles
        }

        try:
            route = self.routes["project_add_member"]
            url = route[1].format(self.endpoint, project, username)
            method = getattr(self, 'do_{0}'.format(route[0]))
            r = method(url, json.dumps(body), "project_add_member", **reqkwargs)
            return AmsUser().load_from_dict(r)
        except AmsException as e:
            raise e

    def get_project_member(self, username, project=None, **reqkwargs):
        """
        Retrieves the respective project member using the provided username with a GET request

        :param username: (str) the username of the user to be retrieved
        :param project: (str) the name of the project the user belongs to.If no
        project is supplied, the declared global project will be used instead
        :param reqkwargs:  keyword arguments that will be passed to underlying
               python-requests library call.
        :return: (AmsUser) the ams user
        """
        try:
            if project is None:
                project = self.project

            route = self.routes["project_get_member"]
            url = route[1].format(self.endpoint, project, username)
            method = getattr(self, 'do_{0}'.format(route[0]))
            r = method(url, "project_get_member", **reqkwargs)
            return AmsUser().load_from_dict(r)
        except AmsException as e:
            raise e

    def create_sub(self, sub, topic, ackdeadline=10, push_endpoint=None,
                   retry_policy_type='linear', retry_policy_period=300, retobj=False, **reqkwargs):
        """This function creates a new subscription in a project with a PUT request

           Args:
               sub: str. The subscription name.
               topic: str. The topic name.
               ackdeadline: int. It is a custom "ack" deadline (in seconds) in
                               the subscription. If your code doesn't
                               acknowledge the message in this time, the
                               message is sent again. If you don't specify
                               the deadline, the default is 10 seconds.
               push_endpoint: URL of remote endpoint that should receive
                              messages in push subscription mode
               retry_policy_type:
               retry_policy_period:
               retobj: Controls whether method should return AmsSubscription object
               reqkwargs: keyword argument that will be passed to underlying
               python-requests library call.
        """
        topic = self.get_topic(topic, retobj=True, **reqkwargs)

        msg_body = json.dumps({"topic": topic.fullname.strip('/'),
                               "ackDeadlineSeconds": ackdeadline})

        route = self.routes["sub_create"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, msg_body, "sub_create", **reqkwargs)

        if push_endpoint:
            ret = self.pushconfig_sub(sub, push_endpoint, retry_policy_type, retry_policy_period, **reqkwargs)
            r['pushConfig'] = {"pushEndpoint": push_endpoint,
                               "retryPolicy": {"type": retry_policy_type,
                                               "period": retry_policy_period}}

        if r['name'] not in self.subs:
            self._create_sub_obj(r, topic.fullname)

        if retobj:
            return self.subs[r['name']]
        else:
            return r

    def delete_sub(self, sub, **reqkwargs):
        """This function deletes a selected subscription in a project

           Args:
               sub: str. The subscription name.
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["sub_delete"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, sub)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, "sub_delete", **reqkwargs)

        sub_fullname = "/projects/{0}/subscriptions/{1}".format(self.project, sub)
        if sub_fullname in self.subs:
            self._delete_sub_obj({'name': sub_fullname})

        return r

    def topic(self, topic, **reqkwargs):
        """Function create a topic in a project.

           It's wrapper around few methods defined in client class. Method will
           ensure that AmsTopic object is returned either by fetching existing
           one or creating a new one in case it doesn't exist.

           Args:
               topic (str): The topic name
           Kwargs:
            reqkwargs: keyword argument that will be passed to underlying
                       python-requests library call.
           Return:
               object (AmsTopic)
        """
        try:
            if self.has_topic(topic, **reqkwargs):
                return self.get_topic(topic, retobj=True, **reqkwargs)
            else:
                return self.create_topic(topic, retobj=True, **reqkwargs)

        except AmsException as e:
            raise e

    def create_topic(self, topic, retobj=False, **reqkwargs):
        """This function creates a topic in a project

           Args:
               topic: str. The topic name.
               retobj: Controls whether method should return AmsTopic object
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["topic_create"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, topic)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, '', "topic_create", **reqkwargs)

        if r['name'] not in self.topics:
            self._create_topic_obj(r)

        if retobj:
            return self.topics[r['name']]
        else:
            return r

    def delete_topic(self, topic, **reqkwargs):
        """This function deletes a topic in a project

           Args:
               topic: str. The topic name.
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """
        route = self.routes["topic_delete"]
        # Compose url
        url = route[1].format(self.endpoint, self.project, topic)
        method = getattr(self, 'do_{0}'.format(route[0]))

        r = method(url, "topic_delete", **reqkwargs)

        topic_fullname = "/projects/{0}/topics/{1}".format(self.project, topic)
        if topic_fullname in self.topics:
            self._delete_topic_obj({'name': topic_fullname})

        return r



import json


class AmsUserProject(object):

    def __init__(self, project="", roles=None, topics=None, subscriptions=None):
        """
        Initialises a new User-Project object.

        NOTE:
        When using a new AmsUserProject object during the creation of a new user in ams,
        you can only affect the project and roles fields.You cannot add a user to a topic
        or subscription through the create functionality.The topics and subscriptions fields
        are present only for view purposes.

        :param project: (str) name of the project
        :param roles: (str[]) roles the user has under the given project
        :param topics: (str[]) topics that the user belongs to under the given project
        :param subscriptions: (str[]) subscriptions that the user belongs to under the given project
        """
        if subscriptions is None or not isinstance(subscriptions, list):
            subscriptions = []
        if topics is None or not isinstance(topics, list):
            topics = []
        if roles is None or not isinstance(roles, list):
            roles = []
        else:
            for r in roles:
                if not isinstance(r, str):
                    raise ValueError("role has to be of type str")
        self.project = project
        self.roles = roles
        self.topics = topics
        self.subscriptions = subscriptions

    def load_from_dict(self, project_dict):
        """
        Accepts a dictionary that contains project data in order to fill the instance's fields
        :param (dict) project_dict: dict that contains project data
        :return: (AmsUserProject) the ams user project object filled with any data provided to its respective fields
        """
        if "project" in project_dict:
            self.project = project_dict["project"]

        if "roles" in project_dict:
            self.roles = project_dict["roles"]
        else:
            self.roles = []

        if "topics" in project_dict:
            self.topics = project_dict["topics"]
        else:
            self.topics = []

        if "subscriptions" in project_dict:
            self.subscriptions = project_dict["subscriptions"]
        else:
            self.subscriptions = []

        return self


class AmsUser(object):

    def __init__(self, uuid="", name="", projects=None, firstname="", lastname="", organization="", description="",
                 token="", email="", service_roles=None, created_on="", created_by="", modified_on=""):
        """

        Initialise a new user object.

        NOTE:
        When using a new AmsUser object during the create functionality, the fields that will affect
        the user creation are name, projects, firstname, lastname, organisation, description, email and
        service_roles.You cannot control the values for the rest of the fields through the create functionality.
        The rest of the fields are present for view purposes.

        :param uuid: (str) the uuid of the user
        :param name: (str) the username
        :param projects: (AmsUserProject[]) list of projects the user belongs to in AMS
        :param firstname: (str) firstname of the user
        :param lastname:  (str) lastname of the user
        :param organization: (str) organisation of the user
        :param description: (str) description of the user
        :param token: (str) user access token for the AMS
        :param email: (str) user email
        :param service_roles: (str[]) possible service roles the user might have in AMS
        :param created_on: (str) user creation date
        :param created_by: (str) user creator
        :param modified_on: (str) timestamp of the latest modification
        """

        if service_roles is None or not isinstance(service_roles, list):
            service_roles = []
        else:
            for s in service_roles:
                if not isinstance(s, str):
                    raise ValueError("service role has to be of type str")

        if projects is None or not isinstance(projects, list):
            projects = []
        else:
            for p in projects:
                if not isinstance(p, AmsUserProject):
                    raise ValueError("project has to be of type AmsUserProject")

        self.uuid = uuid
        self.name = name
        self.projects = projects
        self.firstname = firstname
        self.lastname = lastname
        self.organization = organization
        self.description = description
        self.token = token
        self.email = email
        self.service_roles = service_roles
        self.created_on = created_on
        self.created_by = created_by
        self.modified_on = modified_on

    def to_json(self):
        """
        Utility function that helps convert the AmsUser object to its json string representation
        using json.dumps().
        :return: (str) the json string representation of the object
        """
        user_dict = {}
        if len(self.projects) > 0:
            projects_body = []
            for p in self.projects:
                projects_body.append(
                    {
                        "project": p.project,
                        "roles": p.roles,
                        "topics": p.topics,
                        "subscriptions": p.subscriptions
                    }
                )
            user_dict["projects"] = projects_body

        if self.firstname != "":
            user_dict["first_name"] = self.firstname

        if self.lastname != "":
            user_dict["last_name"] = self.lastname

        if self.description != "":
            user_dict["description"] = self.description

        if self.organization != "":
            user_dict["organization"] = self.organization

        if self.email != "":
            user_dict["email"] = self.email

        if len(self.service_roles) > 0:
            user_dict["service_roles"] = self.service_roles

        return json.dumps(user_dict)

    def load_from_dict(self, user_dict):
        """
        Accepts a dictionary that contains user data in order to fill the instance's fields
        :param (dict) user_dict: dict that contains user data
        :return: (AmsUser) the user object filled with any data provided to its respective fields
        """

        if "uuid" in user_dict:
            self.uuid = user_dict["uuid"]

        if "projects" in user_dict:
            user_projects = []
            for p in user_dict["projects"]:
                user_projects.append(AmsUserProject().load_from_dict(p))
            self.projects = user_projects

        if "name" in user_dict:
            self.name = user_dict["name"]

        if "token" in user_dict:
            self.token = user_dict["token"]

        if "email" in user_dict:
            self.email = user_dict["email"]

        if "service_roles" in user_dict:
            self.service_roles = user_dict["service_roles"]

        if "created_on" in user_dict:
            self.created_on = user_dict["created_on"]

        if "modified_on" in user_dict:
            self.modified_on = user_dict["modified_on"]

        if "created_by" in user_dict:
            self.created_by = user_dict["created_by"]

        if "description" in user_dict:
            self.description = user_dict["description"]

        if "first_name" in user_dict:
            self.firstname = user_dict["first_name"]

        if "last_name" in user_dict:
            self.lastname = user_dict["last_name"]

        if "organization" in user_dict:
            self.organization = user_dict["organization"]

        return self



#!/usr/bin/env python

from argparse import ArgumentParser
from argo_ams_library import ArgoMessagingService, AmsMessage, AmsException

import logging


# setup logger to see the retry messages from ams-library
log = logging.getLogger('argo_ams_library')
log.setLevel(logging.WARNING)
log.addHandler(logging.handlers.SysLogHandler('/dev/log', logging.handlers.SysLogHandler.LOG_USER))
log.addHandler(logging.StreamHandler())


def main():
    parser = ArgumentParser(description="Simple AMS message publish example")
    parser.add_argument('--host', type=str, default='messaging-devel.argo.grnet.gr', help='FQDN of AMS Service')
    parser.add_argument('--token', type=str, required=True, help='Given token')
    parser.add_argument('--project', type=str, required=True, help='Project  registered in AMS Service')
    parser.add_argument('--topic', type=str, required=True, help='Given topic')
    parser.add_argument('--subscription', type=str, required=True, help='Subscription name')
    parser.add_argument('--nummsgs', type=int, default=3, help='Number of messages to pull and ack')
    args = parser.parse_args()

    ams = ArgoMessagingService(endpoint=args.host, token=args.token, project=args.project)

    # static sleep between retry attempts
    msg = AmsMessage(data='foo1', attributes={'bar1': 'baz1'}).dict()
    try:
        ret = ams.publish(args.topic, msg, retry=3, retrysleep=5, timeout=5)
        print(ret)
    except AmsException as e:
        print(e)

    # iptables -A OUTPUT -d messaging-devel.argo.grnet.gr -j DROP

    ackids = list()
    for id, msg in ams.pull_sub(args.subscription, args.nummsgs, retry=3,
                                retrysleep=5, timeout=5):
        data = msg.get_data()
        msgid = msg.get_msgid()
        attr = msg.get_attr()
        print('msgid={0}, data={1}, attr={2}'.format(msgid, data, attr))
        ackids.append(id)

    if ackids:
        ams.ack_sub(args.subscription, ackids, retry=3, retrysleep=5,
                    timeout=5)

    # backoff with each next retry attempt exponentially longer
    msg = AmsMessage(data='foo2', attributes={'bar2': 'baz2'}).dict()
    try:
        ret = ams.publish(args.topic, msg, retry=3, retrybackoff=5, timeout=5)
        print(ret)
    except AmsException as e:
        print(e)

    # iptables -A OUTPUT -d messaging-devel.argo.grnet.gr -j DROP

    ackids = list()
    for id, msg in ams.pull_sub(args.subscription, args.nummsgs,
                                retrybackoff=3, retrysleep=5, timeout=5):
        data = msg.get_data()
        msgid = msg.get_msgid()
        attr = msg.get_attr()
        print('msgid={0}, data={1}, attr={2}'.format(msgid, data, attr))
        ackids.append(id)

    if ackids:
        ams.ack_sub(args.subscription, ackids)

    # static sleep between retry attempts. this example uses consume context
    # method that pull and acks msgs in one call.
    msg = AmsMessage(data='foo3', attributes={'bar3': 'baz3'}).dict()
    try:
        ret = ams.publish(args.topic, msg, retry=3, retrysleep=5, timeout=5)
        print(ret)
    except AmsException as e:
        print(e)

    try:
        msgs = ams.pullack_sub(args.subscription, args.nummsgs, retry=3,
                               retrysleep=5, timeout=5)
        for msg in msgs:
            data = msg.get_data()
            msgid = msg.get_msgid()
            attr = msg.get_attr()
            print('msgid={0}, data={1}, attr={2}'.format(msgid, data, attr))

    except AmsException as e:
        print(e)

main()



from .amsexceptions import AmsException


class AmsTopic(object):
    """Abstraction of AMS Topic

       Topic represents named bucket on AMS service that can hold messages.
       Supported methods are wrappers around the methods defined in client
       class with preconfigured topic name.
    """

    def _build_name(self, fullname):
        return fullname.split('/projects/{0}/topics/'.format(self.init.project))[1]

    def __init__(self, fullname, init):
        self.acls = None
        self.init = init
        self.fullname = fullname
        self.name = self._build_name(self.fullname)

    def delete(self):
        """Delete topic

           Return:
               True: successfull topic deletion
        """

        return self.init.delete_topic(self.name)

    def subscription(self, sub, ackdeadline=10, **reqkwargs):
        """Create a subscription for the topic.

           It's wrapper around few methods defined in client class. Method will
           ensure that AmsSubscription object is returned either by fetching
           existing one or creating a new one in case it doesn't exist.

           Args:
               sub (str): Name of the subscription
           Kwargs:
               ackdeadline (int): Time window in which the AMS service expects
                                  acknowledgement received for pulled messages
               reqkwargs: Keyword arguments that will be passed to underlying
                          python-requests library call.
           Return:
               object (AmsSubscription)

        """
        try:
            if self.init.has_sub(sub, **reqkwargs):
                return self.init.get_sub(sub, retobj=True, **reqkwargs)
            else:
                return self.init.create_sub(sub, self.name, ackdeadline=ackdeadline, retobj=True, **reqkwargs)

        except AmsException as e:
            raise e

    def acl(self, users=None, **reqkwargs):
        """Set or get ACLs assigned to topic

           Kwargs:
               users (list): If list of users is specified, give those user
                             access to topic. Empty list will reset access permission.
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.

        """
        if users is None:
            return self.init.getacl_topic(self.name, **reqkwargs)
        else:
            return self.init.modifyacl_topic(self.name, users, **reqkwargs)

    def iter_subs(self):
        """Generator method that can be used to iterate over subscriptions
           associated with the topic
        """

        for s in self.init.iter_subs(topic=self.name):
            yield s

    def publish(self, msg, retry=0, retrysleep=60, retrybackoff=None, **reqkwargs):
        """Publish message to topic

           Args:
               msg (list, dict): One or list of dictionaries representing AMS
                                 Message
           Kwargs:
               retry: int. Number of request retries before giving up. Default
                           is 0 meaning no further request retry will be made
                           after first unsuccesfull request.
               retrysleep: int. Static number of seconds to sleep before next
                           request attempt
               retrybackoff: int. Backoff factor to apply between each request
                             attempts

               reqkwargs: Keyword argument that will be passed to underlying
                          python-requests library call.

           Returns:
               dict: Dictionary with messageIds of published messages
        """

        return self.init.publish(self.name, msg, retry=retry,
                                 retrysleep=retrysleep,
                                 retrybackoff=retrybackoff, **reqkwargs)



#!/usr/bin/env python

from argparse import ArgumentParser
from argo_ams_library import ArgoMessagingService, AmsException

def main():
    parser = ArgumentParser(description="Simple AMS example of subscription pull/consume")
    parser.add_argument('--host', type=str, default='messaging-devel.argo.grnet.gr', help='FQDN of AMS Service')
    parser.add_argument('--token', type=str, required=True, help='Given token')
    parser.add_argument('--project', type=str, required=True, help='Project  registered in AMS Service')
    parser.add_argument('--subscription', type=str, required=True, help='Subscription name')
    parser.add_argument('--topic', type=str, required=True, help='Given topic')
    parser.add_argument('--nummsgs', type=int, default=3, help='Number of messages to pull and ack')
    args = parser.parse_args()

    # initialize service with given token and project
    ams = ArgoMessagingService(endpoint=args.host, token=args.token, project=args.project)

    # ensure that subscription is created in first run. messages can be
    # pulled from the subscription only when subscription already exists
    # for given topic prior messages being published to topic
    try:
        if not ams.has_sub(args.subscription):
            ams.create_sub(args.subscription, args.topic)
    except AmsException as e:
        print(e)
        raise SystemExit(1)

    # try to pull number of messages from subscription. method will
    # return (ackIds, AmsMessage) tuples from which ackIds and messages
    # payload will be extracted.
    ackids = list()
    for id, msg in ams.pull_sub(args.subscription, args.nummsgs):
        data = msg.get_data()
        msgid = msg.get_msgid()
        attr = msg.get_attr()
        print('msgid={0}, data={1}, attr={2}'.format(msgid, data, attr))
        ackids.append(id)

    # pass list of extracted ackIds to AMS Service so that
    # it can move the offset for the next subscription pull
    # (basically acknowledging pulled messages)
    if ackids:
        ams.ack_sub(args.subscription, ackids)

main()



#!/usr/bin/env python

from argparse import ArgumentParser
from argo_ams_library import ArgoMessagingService, AmsMessage, AmsException

def main():
    parser = ArgumentParser(description="Simple AMS message publish example")
    parser.add_argument('--host', type=str, default='messaging-devel.argo.grnet.gr', help='FQDN of AMS Service')
    parser.add_argument('--token', type=str, required=True, help='Given token')
    parser.add_argument('--project', type=str, required=True, help='Project  registered in AMS Service')
    parser.add_argument('--topic', type=str, required=True, help='Given topic')
    parser.add_argument('--subscription', type=str, required=True, help='Subscription name')
    parser.add_argument('--nummsgs', type=int, default=3, help='Number of messages to pull and ack')
    args = parser.parse_args()

    # initialize service with given token and project
    ams = ArgoMessagingService(endpoint=args.host, token=args.token, project=args.project)

    # ensure that topic is created in first run
    try:
        if not ams.has_topic(args.topic):
            ams.create_topic(args.topic)
    except AmsException as e:
        print(e)
        raise SystemExit(1)

    # publish one message to given topic. message is constructed with
    # help of AmsMessage which accepts data and attributes keys.
    # data is Base64 encoded, attributes is dictionary of arbitrary
    # key/value pairs
    msg = AmsMessage(data='foo1', attributes={'bar1': 'baz1'})
    try:
        ret = ams.publish(args.topic, msg)
        print(ret)
    except AmsException as e:
        print(e)

    # publish a list of two messages to given topic. AmsMessage can also be
    # used as a callable. publish() method accepts either one messages or
    # list of messages
    msg = AmsMessage()
    msglist = [msg(data='foo2', attributes={'bar2': 'baz2'}),
               msg(data='foo3', attributes={'bar3': 'baz3'})]
    try:
        ret = ams.publish(args.topic, msglist)
        print(ret)
    except AmsException as e:
        print(e)

main()



import json

class AmsException(Exception):
    """Base exception class for all Argo Messaging service related errors"""

    def __init__(self, *args, **kwargs):
        super(AmsException, self).__init__(*args, **kwargs)


class AmsServiceException(AmsException):
    """Exception for Argo Messaging Service API errors"""

    def __init__(self, json, request):
        errord = dict()

        self.msg = "While trying the [{0}]: {1}".format(request, json['error']['message'])
        errord.update(error=self.msg)

        if json['error'].get('code'):
            self.code = json['error']['code']
            errord.update(status_code=self.code)

        if json['error'].get('status'):
            self.status = json['error']['status']
            errord.update(status=self.status)

        super(AmsServiceException, self).__init__(errord)


class AmsBalancerException(AmsServiceException):
    """Exception for load balancer Argo Messaging Service errors"""

    def __init__(self, json, request):
        super(AmsBalancerException, self).__init__(json, request)


class AmsTimeoutException(AmsServiceException):
    """Exception for timeouts errors

       Timeouts can be generated by the Argo Messaging Service if message was
       not acknownledged in desired time frame (ackDeadlineSeconds). Also, 408
       timeouts can come from load balancer for partial requests that were not
       completed in required time frame.
    """
    def __init__(self, json, request):
        super(AmsTimeoutException, self).__init__(json, request)


class AmsConnectionException(AmsException):
    """Exception for connection related problems catched from requests library"""

    def __init__(self, exp, request):
        self.msg = "While trying the [{0}]: {1}".format(request, repr(exp))
        super(AmsConnectionException, self).__init__(self.msg)


class AmsMessageException(AmsException):
    """Exception that indicate problems with constructing message"""

    def __init__(self, msg):
        self.msg = msg
        super(AmsMessageException, self).__init__(self.msg)



import sys
import json
import inspect
from base64 import b64encode, b64decode
try:
    from collections.abc import Callable
except ImportError:
    from collections import Callable
from .amsexceptions import AmsMessageException

class AmsMessage(Callable):
    """Abstraction of AMS Message

       AMS Message is constituted from mandatory data field
       and arbitrary number of attributes. Data is Base64
       encoded prior dispatching message to AMS service and
       Base64 decoded when it is being pulled from service.
    """
    def __init__(self, b64enc=True, attributes='', data=None,
                 messageId='', publishTime=''):
        self._attributes = attributes
        self._messageId = messageId
        self._publishTime = publishTime
        self.set_data(data, b64enc)

    def __call__(self, **kwargs):
        if 'attributes' not in kwargs:
            kwargs.update({'attributes': self._attributes})
        self.__init__(b64enc=True, **kwargs)

        return self.dict()

    def _has_dataattr(self):
        if not getattr(self, '_data', False) and not self._attributes:
            raise AmsMessageException('At least data field or one attribute needs to be defined')

        return True

    def set_attr(self, key, value):
        """Set attributes of message

           Args:
               key (str): Key of the attribute
               value (str): Value of the attribute
        """
        self._attributes.update({key: value})

    def set_data(self, data, b64enc=True):
        """Set data of message

           Default behaviour is to Base64 encode data prior sending it by the
           publisher. Method is also used internally when pull from subscription
           is being made by subscriber and in that case, data is already Base64 encoded

           Args:
               data (str): Data of the message
           Kwargs:
               b64enc (bool): Control whether data should be Base64 encoded
        """
        if b64enc and data:
            try:
                if sys.version_info < (3, ):
                    self._data = b64encode(data)
                else:
                    if isinstance(data, bytes):
                        self._data = str(b64encode(data), 'utf-8')
                    else:
                        self._data = str(b64encode(bytearray(data, 'utf-8')), 'utf-8')
            except Exception as e:
                raise AmsMessageException('b64encode() {0}'.format(str(e)))
        elif data:
            self._data = data

    def dict(self):
        """Construct python dict from message"""

        if self._has_dataattr():
            d = dict()
            for attr in ['attributes', 'data', 'messageId', 'publishTime']:
                if getattr(self, '_{0}'.format(attr), False):
                    v = eval('self._{0}'.format(attr))
                    d.update({attr: v})
            return d

    def get_data(self):
        """Fetch the data of the message and Base64 decode it"""

        if self._has_dataattr():
            try:
                return b64decode(self._data)
            except Exception as e:
                raise AmsMessageException('b64decode() {0}'.format(str(e)))

    def get_msgid(self):
        """Fetch the message id of the message"""

        return self._messageId

    def get_publishtime(self):
        """Fetch the publish time of the message set by the AMS service"""

        return self._publishTime

    def get_attr(self):
        """Fetch all attributes of the message"""

        if self._has_dataattr():
            return self._attributes

    def json(self):
        """Return JSON interpretation of the message"""

        return json.dumps(self.dict())

    def __str__(self):
        return str(self.dict())



from setuptools import setup
from os import path
import glob

NAME = 'argo-ams-library'

this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md')) as f:
    long_description = f.read()

def get_ver():
    try:
        with open(NAME + '.spec') as f:
            for line in f:
                if "Version:" in line:
                    return line.split()[1]
    except IOError:
        print("Make sure that %s is in directory" % (NAME + '.spec'))
        raise SystemExit(1)

setup(
    name=NAME,
    version=get_ver(),
    author='SRCE, GRNET',
    author_email='dvrcic@srce.hr, agelos.tsal@gmail.com, kaggis@gmail.com, themiszamani@gmail.com',
    license='ASL 2.0',
    description='A simple python library for interacting with the ARGO Messaging Service',
    long_description=long_description,
    long_description_content_type='text/markdown',
    tests_require=[
        'setuptools_scm',
        'httmock',
        'pytest'
    ],
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.6",
        "Intended Audience :: Developers",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ],
    url='https://github.com/ARGOeu/argo-ams-library',
    package_dir={'argo_ams_library': 'pymod/'},
    packages=['argo_ams_library'],
    install_requires=['requests']
)



class AmsSubscription(object):
    """Abstraction of AMS subscription

       Subscription represents stream of messages that can be pulled from AMS
       service or pushed to some receiver. Supported methods are wrappers
       around methods defined in client class with preconfigured
       subscription name.
    """
    def _build_name(self, fullname):
        return fullname.split('/projects/{0}/subscriptions/'.format(self.init.project))[1]

    def __init__(self, fullname, topic, pushconfig, ackdeadline, init):
        self.acls = None
        self.init = init
        self.fullname = fullname
        self.topic = self.init.topics[topic]
        self.push_endpoint = ''
        self.retry_policy_type = ''
        self.retry_policy_period = ''
        if pushconfig['pushEndpoint']:
            self.push_endpoint = pushconfig['pushEndpoint']
            self.retry_policy_type = pushconfig['retryPolicy']['type']
            self.retry_policy_period = pushconfig['retryPolicy']['period']
        self.ackdeadline = ackdeadline
        self.name = self._build_name(self.fullname)

    def delete(self):
        """Delete subscription"""

        return self.init.delete_sub(self.name)

    def pushconfig(self, push_endpoint=None, retry_policy_type='linear',
                   retry_policy_period=300, **reqkwargs):
        """Configure Push mode parameters of subscription. When push_endpoint
           is defined, subscription will automatically start to send messages to it.

           Kwargs:
               push_endpoint (str): URL of remote endpoint that will receive messages
               retry_policy_type (str):
               retry_policy_period (int):
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.

        """

        return self.init.pushconfig_sub(self.name, push_endpoint=push_endpoint,
                                        retry_policy_type=retry_policy_type,
                                        retry_policy_period=retry_policy_period,
                                        **reqkwargs)

    def pull(self, num=1, retry=0, retrysleep=60, retrybackoff=None,
             return_immediately=False, **reqkwargs):
        """Pull messages from subscription

           Kwargs:
               num (int): Number of messages to pull
               retry: int. Number of request retries before giving up. Default
                           is 0 meaning no further request retry will be made
                           after first unsuccesfull request.
               retrysleep: int. Static number of seconds to sleep before next
                           request attempt
               retrybackoff: int. Backoff factor to apply between each request
                             attempts
               return_immediately (boolean): If True and if stream of messages is empty,
                                             subscriber call will not block and wait for
                                             messages
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
           Return:
               [(ackId, AmsMessage)]: List of tuples with ackId and AmsMessage instance
        """

        return self.init.pull_sub(self.name, num=num,
                                  return_immediately=return_immediately,
                                  **reqkwargs)

    def pullack(self, num=1, retry=0, retrysleep=60, retrybackoff=None,
                return_immediately=False, **reqkwargs):
        """Pull messages from subscription and acknownledge them in one call.

          If succesfull subscription pull immediately follows with failed
          acknownledgment (e.g. network hiccup just before acknowledgement of
          received messages), consume cycle will reset and start from
          begginning with new subscription pull.

           Kwargs:
               num (int): Number of messages to pull
               retry: int. Number of request retries before giving up. Default
                           is 0 meaning no further request retry will be made
                           after first unsuccesfull request.
               retrysleep: int. Static number of seconds to sleep before next
                           request attempt
               retrybackoff: int. Backoff factor to apply between each request
                             attempts
               return_immediately (boolean): If True and if stream of messages is empty,
                                             subscriber call will not block and wait for
                                             messages
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
           Return:
               [AmsMessage1, AmsMessage2]: List of AmsMessage instances
        """

        return self.init.pullack_sub(self.name, num=num,
                                     return_immediately=return_immediately,
                                     **reqkwargs)

    def time_to_offset(self, timestamp, **reqkwargs):
        """
           Retrieve the closest(greater than) available offset to the given timestamp.

           Args:
               timestamp(datetime.datetime): The timestamp of the offset we are looking for.

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
        """

        return self.init.time_to_offset_sub(self.name, timestamp, **reqkwargs)

    def offsets(self, offset='all', move_to=0, **reqkwargs):
        """
           Retrieve the positions of min, max and current offsets or move
           current offset to new one.

           Args:
               offset (str): The name of the offset. If not specified, it will
                             return all three of them as a dict. Values that can
                             be specified are 'max', 'min', 'current' and 'all'.
               move_to (int): New position for current offset.

           Kwargs:
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.
           Return:
                 dict: A dictionary containing all 3 offsets. If move_to
                       is specified, current offset will be moved and updated.
                 int: The value of the specified offset.
        """
        avail_offsets = set(['max', 'min', 'current'])

        if (offset == 'all' or offset in avail_offsets) and move_to == 0:
            return self.init.getoffsets_sub(self.name, offset, **reqkwargs)
        elif move_to != 0:
            _ = self.init.modifyoffset_sub(self.name, move_to, **reqkwargs)
            return self.init.getoffsets_sub(self.name, offset, **reqkwargs)

    def acl(self, users=None, **reqkwargs):
        """Set or get ACLs assigned to subscription

           Kwargs:
               users (list): If list of users is specified, give those user
                             access to subscription. Empty list will reset access permission.
               reqkwargs: keyword argument that will be passed to underlying
                          python-requests library call.

        """
        if users is None:
            return self.init.getacl_sub(self.name, **reqkwargs)
        else:
            return self.init.modifyacl_sub(self.name, users, **reqkwargs)

    def ack(self, ids, retry=0, retrysleep=60, retrybackoff=None, **reqkwargs):
        """Acknowledge receive of messages

           Kwargs:
               ids (list): A list of ackIds of the messages to acknowledge.
               retry: int. Number of request retries before giving up. Default
                           is 0 meaning no further request retry will be made
                           after first unsuccesfull request.
               retrysleep: int. Static number of seconds to sleep before next
                           request attempt
               retrybackoff: int. Backoff factor to apply between each request
                             attempts
        """

        return self.init.ack_sub(self.name, ids, **reqkwargs)



import logging

try:
    from logging import NullHandler
except ImportError:
    class NullHandler(logging.Handler):
        def emit(self, record):
            pass
logging.getLogger(__name__).addHandler(NullHandler())

from .ams import ArgoMessagingService
from .amsexceptions import (AmsServiceException, AmsBalancerException,
                            AmsConnectionException, AmsTimeoutException,
                            AmsMessageException, AmsException)
from .amsmsg import AmsMessage
from .amstopic import AmsTopic
from .amssubscription import AmsSubscription
from .amsuser import AmsUser, AmsUserProject
